@@ -56,7 +56,7 @@ class Tuple2(click.ParamType):
 
 @click.group()
 @click.option("-p", "--parallel", default=1, help="Run with this number of parallel processes. If 0, use number of cores.")
-@click.version_option(version="4.1.2")
+@click.version_option(version="4.2.0")
 @click.pass_context
 def main(ctx, parallel):
   """
@@ -555,9 +555,12 @@ def mesh_forge(
 @click.argument("path")
 @click.option('--queue', required=True, help="AWS SQS queue or directory to be used for a task queue. e.g. sqs://my-queue or ./my-queue. See https://github.com/seung-lab/python-task-queue", type=str)
 @click.option('--magnitude', default=2, help="Split up the work with 10^(magnitude) prefix based tasks. Default: 2 (100 tasks)")
+@click.option('--nlod', default=0, help="(multires) How many extra levels of detail to create.", show_default=True)
+@click.option('--vqb', default=16, help="(multires) Vertex quantization bits for stored model representation. 10 or 16 only.", show_default=True)
+@click.option('--min-chunk-size', type=Tuple3(), default="512,512,512",  help="(multires) Sets the minimum chunk size of the highest resolution mesh fragment.", show_default=True)
 @click.option('--dir', default=None, help="Write manifests into this directory instead of the one indicated in the info file.")
 @click.pass_context
-def mesh_merge(ctx, path, queue, magnitude, dir):
+def mesh_merge(ctx, path, queue, magnitude, nlod, vqb, dir, min_chunk_size):
   """
   (2) Merge the mesh pieces produced from the forging step.
 
@@ -567,9 +570,18 @@ def mesh_merge(ctx, path, queue, magnitude, dir):
   file that is an index for locating the fragments.
   """
   path = cloudfiles.paths.normalize(path)
-  tasks = tc.create_mesh_manifest_tasks(
-    path, magnitude=magnitude, mesh_dir=dir
-  )
+
+  if nlod > 0:
+    tasks = tc.create_unsharded_multires_mesh_tasks(
+      path, num_lod=nlod, 
+      magnitude=magnitude, mesh_dir=dir,
+      vertex_quantization_bits=vqb,
+      min_chunk_size=min_chunk_size,
+    )
+  else:
+    tasks = tc.create_mesh_manifest_tasks(
+      path, magnitude=magnitude, mesh_dir=dir
+    )
 
   parallel = int(ctx.obj.get("parallel", 1))
   tq = TaskQueue(normalize_path(queue))
@@ -578,6 +590,7 @@ def mesh_merge(ctx, path, queue, magnitude, dir):
 @meshgroup.command("merge-sharded")
 @click.argument("path")
 @click.option('--queue', required=True, help="AWS SQS queue or directory to be used for a task queue. e.g. sqs://my-queue or ./my-queue. See https://github.com/seung-lab/python-task-queue", type=str)
+@click.option('--nlod', default=1, help="Number of levels of detail to create.", type=int, show_default=True)
 @click.option('--vqb', default=16, help="Vertex quantization bits. Can be 10 or 16.", type=int, show_default=True)
 @click.option('--compress-level', default=7, help="Draco compression level.", type=int, show_default=True)
 @click.option('--shard-index-bytes', default=2**13, help="Size in bytes to make the shard index.", type=int, show_default=True)
@@ -585,12 +598,13 @@ def mesh_merge(ctx, path, queue, magnitude, dir):
 @click.option('--min-shards', default=1, help="Minimum number of shards to generate. Excess shards can make it easier to parallelize the merge process.", type=int, show_default=True)
 @click.option('--minishard-index-encoding', default="gzip", help="Minishard indices can be compressed. gzip or raw.", show_default=True)
 @click.option('--spatial-index-db', default=None, help="CloudVolume generated SQL database for spatial index.", show_default=True)
+@click.option('--min-chunk-size', type=Tuple3(), default="512,512,512",  help="(multires) Sets the minimum chunk size of the highest resolution mesh fragment.", show_default=True)
 @click.pass_context
 def mesh_sharded_merge(
   ctx, path, queue, 
-  vqb, compress_level,
+  nlod, vqb, compress_level,
   shard_index_bytes, minishard_index_bytes, min_shards,
-  minishard_index_encoding, spatial_index_db
+  minishard_index_encoding, spatial_index_db, min_chunk_size
 ):
   """
   (2) Postprocess fragments into finished sharded multires meshes.
@@ -607,6 +621,7 @@ def mesh_sharded_merge(
   path = cloudfiles.paths.normalize(path)
   tasks = tc.create_sharded_multires_mesh_tasks(
     path, 
+    num_lod=nlod,
     draco_compression_level=compress_level,
     vertex_quantization_bits=vqb,
     shard_index_bytes=shard_index_bytes,
@@ -614,6 +629,7 @@ def mesh_sharded_merge(
     minishard_index_encoding=minishard_index_encoding,
     min_shards=min_shards,
     spatial_index_db=spatial_index_db,
+    min_chunk_size=min_chunk_size,
   )
 
   parallel = int(ctx.obj.get("parallel", 1))
